/*
 *  savecore.c
 *
 *  Copyright (C) 2004  FUJITSU LIMITED
 *  Copyright (C) 2001-2004  Red Hat, Inc
 *
 *  Written by Nobuhiro Tachino (tachino@jp.fujitsu.com) based on
 *  netdump-server code.
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/* $Id: savecore.c,v 1.1.1.1 2004/06/01 19:02:10 anderson Exp $ */

#include <glib.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/statfs.h>
#include <errno.h>
#include <string.h>
#include <stdint.h>
#include <time.h>

#include "diskdumpelf.h"
#include "dumpdevice.h"

#define Err(x, ...) fprintf(stderr, "savecore: " x "\n", ## __VA_ARGS__)
#define Syserr(x, ...) fprintf(stderr, "savecore: " x ": %s\n", ## __VA_ARGS__, strerror(errno))

int verbose = 0;

/*
 * We save the disk dump under /var/crash/127.0.0.1-<date>/..
 */
#define DISKDUMP_DIR_PREFIX "/var/crash"
#define DISKDUMP_IP	0x7f000001

#define roundup(x, y)  ((((x)+((y)-1))/(y))*(y))

static DumpDevice *
create_new_device(const char *device_name)
{
	DumpDevice *device;

	device = g_new0(DumpDevice, 1);

	device->device_name = g_strdup(device_name);

	device->file_fd = -1;
	device->device_fd = -1;

	/* Lame default. should be read from kernel */
	device->page_offset = 0xC0000000;

	return device;
}

static int
open_device_file(DumpDevice *device)
{
	int fd;

	fd = open(device->device_name, O_RDONLY);
	if (fd < 0) {
		Syserr("unable to open device file %s", device->device_name);
		return -1;
	}
	device->device_fd = fd;
	return 0;
}

static int
execute_script(const char *script, const char *dir)
{
	struct stat sbuf;
	char *filename;
	char *cmdline;
	int res = -1;

	filename = g_strconcat(DISKDUMP_DIR_PREFIX "/scripts/",
			  script,
			  NULL);

	if (stat(filename, &sbuf) == 0) {
		cmdline = g_strdup_printf("%s '%s'",
					 filename, (dir)?dir:"");
		res = system(cmdline);
		g_free(cmdline);
	}

	g_free(filename);
	return res;
}

static int check_fs(DumpDevice *device)
{
	struct statfs buf;
	struct disk_dump_header *header = device->header;
	off_t memdump_size;
	int ret;
	gboolean tried_once = FALSE;

redo:
	if (fstatfs(device->file_fd, &buf) < 0) {
		Err("unable to statfs on the crash dump file");
		return -1;
	}

	memdump_size = (off_t)header->written_blocks + 3; /* elf header */
	memdump_size *= header->block_size;
	memdump_size = (memdump_size + (buf.f_bsize - 1)) / buf.f_bsize;

	if (memdump_size <= buf.f_bavail)
		return 0;
	if (!tried_once++) {
		ret = execute_script("diskdump-nospace", device->dump_dir);
		if (ret > 0)
			goto redo;
	}

	Err("no space for dump image");
	return -1;
}

static const char *
get_dir(DumpDevice *device)
{
	char *path;
	time_t timeofdeath;
	int res;
	guint32 ip = DISKDUMP_IP;
	struct tm *tm;
	struct disk_dump_header *header = device->header;
	char date[128];

	timeofdeath = (time_t)header->timestamp.tv_sec;

	tm = localtime(&timeofdeath);
	strftime(date, sizeof(date), "%Y-%m-%d-%H:%M", tm);
	path = g_strdup_printf("%s/%d.%d.%d.%d-%s",
			       DISKDUMP_DIR_PREFIX,
			       (ip >> 24) & 0xff,
			       (ip >> 16) & 0xff,
			       (ip >> 8) & 0xff,
			       (ip) & 0xff,
			       date);

	res = mkdir(path, S_IRWXU);
	if (res) {
		Err("error making crashdump directory %s", path);
		g_free(path);
		return NULL;
	}

	return path;
}

static int
open_dump_file(DumpDevice *device)
{
	const char *dir;
	char *file_name;
	int fd;

	dir = get_dir(device);
	if (dir == NULL)
		return -1;

	device->dump_dir = dir;

	file_name = g_strconcat(device->dump_dir, "/vmcore-incomplete", NULL);
	fd = open(file_name, O_WRONLY|O_TRUNC|O_CREAT, S_IRUSR|S_IWUSR);
	if (fd < 0) {
		Syserr("unable to open crash dump file %s", file_name);
		g_free(file_name);
		return -1;
	}

	device->file_fd		= fd;
	device->file_name	= file_name;

	return 0;
}

static inline off_t header_offset(struct disk_dump_header *header)
{
	return (off_t)header->block_size;
}

static inline int sub_header_offset(struct disk_dump_header *header)
{
	return (off_t)header->block_size + header_offset(header);
}

static int
read_device_header(DumpDevice *device)
{
	struct disk_dump_header *header;
	int ret;
	int bitmap_len;
	int block_size;

	block_size = (int)sysconf(_SC_PAGESIZE);

	header = g_malloc(block_size);

	lseek(device->device_fd, block_size, SEEK_SET);

	ret = read(device->device_fd, header, block_size);
	if (ret < 0 || ret != block_size) {
		Err("cannot read dump header");
		return -1;
	}

	/* validate dump header */
	if (memcmp(header->signature, DISK_DUMP_SIGNATURE,
				sizeof(header->signature))) {
		Err("device does not have panic dump header");
		return -1;
	}
	if (header->block_size != block_size) {
		Err("block size in the dump header does not match with system page size");
		return -1;
	}
	if (sizeof(*header) + sizeof(void *) * header->nr_cpus > block_size ||
	    header->nr_cpus <= 0) {
		Err("invalid nr_cpus value: %d", header->nr_cpus);
		return -1;
	}

	/* read sub header */
	lseek(device->device_fd, block_size * 2, SEEK_SET);
	device->regs = g_malloc(sizeof(elf_gregset_t));
	ret = read(device->device_fd, device->regs, sizeof(elf_gregset_t));
	if (ret < 0 || ret != sizeof(elf_gregset_t)) {
		Err("cannot read dump sub header");
		return -1;
	}

	/* read memory bitmap */
	bitmap_len = block_size * header->bitmap_blocks;

	lseek(device->device_fd, block_size * (2 + header->sub_hdr_size), SEEK_SET);
	device->bitmap = g_malloc(bitmap_len);
	ret = read(device->device_fd, device->bitmap, bitmap_len);
	if (ret < 0 || ret != bitmap_len) {
		Err("cannot read memory bitmap");
		return -1;
	}


	/*
	 * page 0: device header
	 * page 1: dump header
	 * page 2..: dump sub header
	 * page 2 + size of subheader: memory page 0
	 */
	device->device_page_start = (2 + header->sub_hdr_size + header->bitmap_blocks) * header->block_size;

	device->header = header;

	return 0;
}

/******************************************************************************
 *                       Elf core dumping                                     *
 ******************************************************************************/

static size_t
dump_elf32_note(char *buf, Elf32_Word type, char *name, char *desc, int d_len)
{
	Elf32_Nhdr *note;
	size_t len;

	note = (Elf32_Nhdr *)buf;
	note->n_namesz = strlen(name);
	note->n_descsz = d_len;
	note->n_type = type;
	len = sizeof(Elf32_Nhdr);

	memcpy(buf + len, name, note->n_namesz);
	len = roundup(len + note->n_namesz, 4);

	memcpy(buf + len, desc, note->n_descsz);
	len = roundup(len + note->n_descsz, 4);

	return len;
}


static gboolean
generate_i386_elf_header(DumpDevice *device)
{
	char *buffer, *ptr;
	Elf32_Ehdr *elf;
	Elf32_Phdr *notes;
	Elf32_Phdr *load;
	size_t offset, len;
	void *taskstruct[2];
	struct elf_prpsinfo prpsinfo;
	struct elf_prstatus prstatus;
	struct disk_dump_header *header = device->header;

	buffer = g_malloc0(sizeof(Elf32_Ehdr) + 2 * sizeof (Elf32_Phdr) +
		4096*2); /* 4096*2 should be enought for the notes + roundup */

	offset = 0;
	ptr = buffer;

	/* Elf header */
	elf = (Elf32_Ehdr *)ptr;
	memcpy (elf->e_ident, ELFMAG, SELFMAG);
	elf->e_ident[EI_CLASS] = ELFCLASS32;
	elf->e_ident[EI_DATA]	= ELFDATA2LSB;
	elf->e_ident[EI_VERSION]= EV_CURRENT;
	memset (elf->e_ident+EI_PAD, 0, EI_NIDENT-EI_PAD);

	elf->e_type	= ET_CORE;
	elf->e_machine  = EM_386;
	elf->e_version  = EV_CURRENT;
	elf->e_entry	= 0;
	elf->e_phoff	= sizeof (Elf32_Ehdr);
	elf->e_shoff	= 0;
	elf->e_flags	= 0;
	elf->e_ehsize	= sizeof (Elf32_Ehdr);
	elf->e_phentsize = sizeof (Elf32_Phdr);
	elf->e_phnum	= 2;
	elf->e_shentsize = 0;
	elf->e_shnum	= 0;
	elf->e_shstrndx	= 0;

	offset += sizeof (Elf32_Ehdr);
	ptr += sizeof (Elf32_Ehdr);

	/* PT_NOTE */

	notes = (Elf32_Phdr *)ptr;
	notes->p_type	  = PT_NOTE;
	notes->p_offset = 0; /* TO BE FILLED IN */
	notes->p_vaddr  = 0;
	notes->p_paddr  = 0;
	notes->p_filesz = 0; /* TO BE FILLED IN */
	notes->p_memsz  = 0;
	notes->p_flags  = 0;
	notes->p_align  = 0;

	offset += sizeof(Elf32_Phdr);
	ptr += sizeof(Elf32_Phdr);

	/* PT_LOAD */

	load = (Elf32_Phdr *)ptr;
	load->p_type   = PT_LOAD;
	load->p_offset = 0; /* TO BE FILLED IN */
	load->p_vaddr  = device->page_offset;
	load->p_paddr  = 0;
	load->p_filesz = header->block_size * header->max_mapnr;
	load->p_memsz  = load->p_filesz;
	load->p_flags  = PF_R|PF_W|PF_X;
	load->p_align  = header->block_size;

	offset += sizeof(Elf32_Phdr);
	ptr += sizeof(Elf32_Phdr);

	notes->p_offset = offset;

	/* NT_PRSTATUS note */
	if (device->regs)
	memcpy(&prstatus.pr_reg, device->regs, sizeof(prstatus.pr_reg));
		len = dump_elf32_note(ptr, NT_PRSTATUS, "CORE",
		(char *)&prstatus, sizeof(prstatus));
	offset += len;
	ptr += len;
	notes->p_filesz += len;


	/* NT_PRPSINFO note */
	memset(&prpsinfo, 0, sizeof (struct elf_prpsinfo));
	prpsinfo.pr_state	= 0;
	prpsinfo.pr_sname	= 'R';
	prpsinfo.pr_zomb	= 0;
	strcpy(prpsinfo.pr_fname, "vmlinux");

	len = dump_elf32_note(ptr, NT_PRPSINFO, "CORE",
		(char *)&prpsinfo, sizeof(prpsinfo));

	offset += len;
	ptr += len;
	notes->p_filesz += len;

	/*
	 * fill NT_TASKSTRUCT note with the addresses of all task_structs
	 * XXX It may be helpfull to add current_cpu too.
	 */
	taskstruct[0] = header->tasks[header->current_cpu];
	taskstruct[1] = NULL;	/* dummy switch_stack */

	len = dump_elf32_note(ptr, NT_TASKSTRUCT, "CORE",
			(char *)taskstruct, sizeof(taskstruct));
	offset += len;
	ptr += len;
	notes->p_filesz += len;

	/* Page align offset */
	offset = roundup(offset, header->block_size);

	load->p_offset = offset;
	device->data_offset = offset;

	while (offset > 0) {
		len = write(device->file_fd,
			buffer + (device->data_offset - offset), offset);
		if (len < 0) {
			g_free(buffer);
			return TRUE;
		}

		offset -= len;
	}

	g_free(buffer);
	return FALSE;
}

static int
write_dump_header(DumpDevice *device)
{
	if (generate_i386_elf_header(device))
		return -1;
	else
		return 0;
}

static inline gboolean
check_zero_page(DumpDevice *device, char *data)
{
	int i;

	for (i = 0; i < device->header->block_size; i++)
		if (data[i])
			return FALSE;
	return TRUE;
}

static inline int page_is_ram(int nr, DumpDevice *device)
{
	return device->bitmap[nr>>3] & (1 << (nr & 7));
}

static int
write_pages(DumpDevice *device)
{
	int len;
	int i;
	char *data;
	int ret = 0;
	struct disk_dump_header *header = device->header;
	int block_size = header->block_size;

	data = g_malloc(block_size);

	lseek(device->device_fd, (off_t)device->device_page_start, SEEK_SET);
	lseek(device->file_fd, (off_t)device->data_offset, SEEK_SET);

	for (i = 0; i < header->max_mapnr; i++) {
		if (page_is_ram(i, device)) {
			len = read(device->device_fd, data, block_size);
			if (len != block_size) {
				Err("device read failed");
				ret = -1;
				goto out;
			}
			if (check_zero_page(device, data)) {
				lseek(device->file_fd, block_size, SEEK_CUR);
				continue;
			}
			len = write(device->file_fd, data, block_size);
			if (len != block_size) {
				Syserr("error writing memory dump");
				ret = -1;
				goto out;
			}
		} else {
			lseek(device->file_fd, block_size, SEEK_CUR);
		}
	}

	/* 
	 * Enforce to set the size of vmcore because last block may not be
	 * actually written.
	 */
	ret = ftruncate(device->file_fd, (off_t)device->data_offset +
					 (off_t)header->max_mapnr * block_size);
	if (ret < 0) {
		perror("ftruncate");
		goto out;
	}
out:
	g_free(data);
	return ret;
}

static int
rename_vmcore(DumpDevice *device)
{
	char *old_file;
	char *new_file;

	old_file = g_strconcat(device->dump_dir, "/vmcore-incomplete", NULL);
	new_file = g_strconcat(device->dump_dir, "/vmcore", NULL);

	if (link(old_file, new_file) == 0)
		unlink(old_file);

	g_free(old_file);
	g_free(new_file);

	return 0;
}

static void
usage(void) {
	fprintf(stderr, "Usage: savecore [-v] device\n");
	exit(1);
}

int
main(int argc, char *argv[])
{
	DumpDevice *device;
	struct disk_dump_header *header;
	int opt;
	int ret;
	unsigned int total_blocks;

	while ((opt = getopt(argc, argv, "v")) != -1) {
		switch (opt) {
		case 'v':
			verbose =1;
			break;
		default:
			usage();
			break;
		}
	}

	if (optind + 1 != argc)
		usage();

	device = create_new_device(argv[optind]);
	if (device == NULL)
		exit(1);
	if ((ret = open_device_file(device)) < 0)
		exit(1);
	if ((ret = read_device_header(device)) < 0)
		exit(1);
	if ((ret = open_dump_file(device)) < 0)
		exit(1);
	if ((ret = check_fs(device)) < 0)
		exit(1);
	if ((ret = write_dump_header(device)) < 0)
		exit(1);
	if ((ret = write_pages(device)) < 0)
		exit(1);

	header = device->header;

	total_blocks =    1 /* partition header */
			+ 1 /* common header */
		        + header->sub_hdr_size
			+ header->bitmap_blocks
			+ header->total_ram_blocks;

	if (total_blocks != header->written_blocks) {
		Err("dump partition has only %u pages. %u expected",
			header->written_blocks, header->total_ram_blocks);
		exit(1);
	} else {
		if ((ret = rename_vmcore(device)) != 0)
			exit(1);
	}

	exit(0);
}
